
/*==================================================================================

 

  cpNode - Control Point CMRI Node

  Version 2.0

 

  ------------------------------------------------------------------------------------

  cpNode concept and design committed on 5/31/2013 by Chuck Catania and Seth Neumann

 

  This work is licensed under the Creative Commons Attribution-ShareAlike 3.0 Unported License.

  To view a copy of the license, visit http://creativecommons.org/licenses/by-sa/3.0/deed.en_US

  ------------------------------------------------------------------------------------

 

  Author: Chuck Catania - Model Railroad Control Systems

  System Board:  Modern Devices BBLeo (Arduino Style Leonardo)

 

  Revision History:

   v2.00  04/01/2020 (Andrew Deak) Add DCC to Mega and DCC Assesory Decoder-I2C

   v1.60  01/18/2018 (Andrew Deak) Add 16IN40OUT to Arduino Mega support

   v1.5.5 05/25/2017 (Andrew Deak) Changed Pin Names to work with Ardunio 1.87

   v1.4.2 05/27/2015  (TVerberg) Corrected typo errors in Base_Node_12out_4in output unpacking,

                      (TVerberg) Corrected Base_Node_Servo input packing routines.

                      Rearranged CMRInet protocol responses to put non-message functions at the front of the list.

                      Added Init message DL/DH delay processing for Classic node compatability.

                      Added Init_CA_Ports_OFF boolean.  If set to true, all ports driving Common Anode LEDs will be forced OFF at bootup.

                      Changed CMRInet_BufSize to int and increased length to 260 to handle the max SUSIC + 4 pad buffer

                      Changed Flush_CMRInet_To_ETX() to exit on either seeing ETX or empty serial buffer

                      Changed DEBOUNCE_DELAY from 10 ms to 2 ms

                      Added BASE_NODE_RSMC_LOCK configuration to support locking an RSMC controlled turnout with

                       Dennis Drury's switch lock board.

                      Added BASE_NODE_8OUT8IN.  Sets D4-D11 as outputs, D12-A5 as inputs.

   v1.4.1 04/15/2015  Changed CMRINET_SPEED definition from int to long for network speeds greater than 28800 bps

   v1.4   06/25/2014  Added the 12 output, 4 input standard configuration per Dick Johannes of the NMRA HUB Division

                      Moved debug option and SN variables out of Node Configuration Parameters area.

   v1.3   04/06/2014  Fixed issue with BASE_NODE_8IN8OUT, BASE_NODE_16OUT high bit B8 output not moved to A5.

                      Bit extraction loop ended one bit early.  Other routines worked because loop limit was less than

                      maximum port map index.

   v1.2   03/06/2014  Fixed issue with BASE_NODE_8IN8OUT where port setup did not match specification.

                      This was an implementation deviation from the design specification.

   v1.1   03/01/2014  Fixed issue with BASE_NODE_8IN8OUT where the byte assignment was flipped.

   v1.0   01/04/2014  Released

    0.0d  08/24/2013  Initial template definition

 

  This sketch is the code template for a CMRI serial protocol node implemented in an Arduino style system board.

  The Modern Devices BBLeo, Bare Bones Leonardo (ATMega32u4) is the target Arduino style system board.

 

  Implements the CMRI Serial Protocol designed by Dr. Bruce Chubb and published publicly in various books, magazines, and articles.

  The physical link is RS485, 4-wire, half duplex, serial.

  Each node has a one byte address in the range of 0-127.

  The Host polls the CMRI nodes for data.

  Port assignments represent the defined capability of a cpNode.

  The base node data configuration is two bytes in, two bytes out.

  All needed signal pins are connected via pin headers to the node board.

 

  Reserved I/O pins are:

   D0 - RX  CMRI RS485 Receive

   D1 - TX  CMRI RS485 Transmit

   D2 - SDA I2C Data

   D3 - SCL I2C Clock

 

                   SERIAL PORT ASSIGNMENTS

 

  Serial        Bootloader/Debug monitor (USB Built in)

  Serial1       CMRI Port (RS-485) RX(D0) and TX(D1) pins

 

  /**************************************************************/

/**********        CMRI Protocol Message Format      **********/

/**************************************************************

  - Initialization Message (I)  HOST to NODE

  SYN SYN STX <UA> <I><NDP><dH><dL><NS><CT(1)><CT(NS)> ETX

 

  - Poll for Data (P)  HOST to NODE

  SYN SYN STX <UA> <P> ETX

 

  - Read Data (R)  NODE To HOST  (Response to Poll)

  SYN SYN STX <UA> <R><IB(1)><IB(NS)> ETX

 

  - Transmit Data (T)  HOST to NODE

  SYN SYN STX <UA> <T><OB(1)><OB(NS)> ETX

*/

 

/*************************************************************/

/*********            Library Header Files           *********/

/*************************************************************/

#include <SoftwareSerial.h>

#include <Wire.h>

#include <NmraDcc.h>

 

 

 

 

// 9680 Bytes BASE_NODE_8IN8OUT

// 11266 Bytes with IOX

 

//#include <Servo.h>

// 11798 Bytes

 

/*==============================================*/

/*====    NODE CONFIGURATION PARAMETERS     ====*/

/*==============================================*/

//

int nodeID =  0;                                //

 

const char* NodeIP = " 192.168.1.50";

 

const char* NodeType = "  CMRIOverIP ";//

 

const long CMRINET_SPEED =  19200;               //

//

boolean Init_CA_Ports_OFF = false;              //

//

/*==============================================*/

/*====    NODE CONFIGURATION PARAMETERS     ====*/

/*==============================================*/

 

//--------------------------------------------------------

//  Uncomment only one of the standard node configurations

//  #ifdef selects the higher order bit definitions

//-------------------------------------------------------

//#define BASE_NODE

//#define BASE_NODE_8IN8OUT

//#define BASE_NODE_8OUT8IN    // cpNode System Test Configuration

//#define BASE_NODE_12OUT4IN

//#define BASE_NODE_16IN

//#define BASE_NODE_16OUT

//#define BASE_NODE_RSMC

//#define BASE_NODE_RSMC_LOCK

#define BASE_NODE_16IN40OUT   //Arduino Mega

//----------------------

// I/O Expander boards

//----------------------

#define USE_IOX

//----------------------

// DCC_Accessory_Decoder

//----------------------

#define DCC_Accessory_Decoder

 

 

#ifdef DCC_Accessory_Decoder

// DCC Define

NmraDcc  Dcc ;

DCC_MSG  Packet ;

uint16_t Accessory_Address  =  1;           // THIS ADDRESS IS THE START OF THE SWITCHES RANGE using Ardunio Mega

uint16_t Accessory_Address_I2C  =  101;           // THIS ADDRESS IS THE START OF THE SWITCHES RANGE using the I2C Boards

 

// Define Mega DCC Assesory Outputs Pins

int DCC_Turnout_Mega_Max = 16;

// Use D3-D48 and or A0-A15

int DCC_Turnout_Pins [] = { A8, A9, A10, A11, A12, A13, A14, A15};

 

//DCC I/O controled by DCC I2C assy decoader

uint16_t   Addr_DCC;

// Set DCC_Turnout_Pins_I2C to Pins control by DCC and address will be the Accessory_Address_I2C the pin

uint16_t  DCC_Turnout_Pins_I2C [] = {3, 11, 22, 26, 34, 41, 51, 59, 70, 76 ,82, 91,98,106,116,122};

 

//******DCC Pins = DCC Pins + Accessory_Address_I2C *******

//Set Pins to 0 or 1 to make pin on or off (default)

//                         Ox20 A      Ox20 B      Ox21 A     Ox21 B      Ox22 A      Ox22 B      Ox23 A      Ox23 B       Ox24 A     Ox24 B      Ox25 A     Ox25 B      Ox26 A       Ox26 B      Ox27 A      Ox27 B

//        MegaI/O Board     J11         J12         J13        J14         J15         J16         J17         J18         J11         J12         J13         J14         J15         J16         J17         J18

//           Pins        87654321    87654321    87654321    87654321    87654321    87654321    87654321    87654321    87654321    87654321    87654321    87654321    87654321    87654321    87654321    87654321

//         DCC Pins      8------1    16-----9    24----17    32----25    40----33    48----41    54----49    64----55    72----65    80----73    88----81    96----89    104---97    112--105    120--113    128--121

uint16_t  DCC_I2C[] = {0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000};

 

 

 

#endif

 

/************************************************/

/*****        STANDARD DATA PORT MAP        *****/

/************************************************/

const int

// RS-485 Port

//------------

RX      =  0,    // D0  CMRI RS485 Receive

TX      =  1,    // D1  CMRI RS485 Transmit

 

// I2C Port

//---------

i2C_SDA =  SDA,  // D20  SDA (data)

i2c_SCL =  SCL,  // D21  SCL (clock)

 

// Digital Pins

//-------------

//IB(0) or OB(0)

D4     =  4,    // D4

D5     =  5,    // D5

D6     =  6,    // D6

D7     =  7,    // D7

D8     =  8,    // D8

D9     =  9,    // D9

D10    = 10,    // D10

D11    = 11,    // D11

//IB(1) or OB(1)Plus A0-A5

D12    = 12,    // D12

D13    = 13,    // D13

//IB(2) or OB(2)

D22     =  22,    // D22

D23     =  23,    // D23

D24     =  24,    // D24

D25     =  25,    // D25

D26     =  26,    // D26

D27     =  27,    // D27

D28     =  28,    // D28

D29     =  29,    // D29

//IB(3) or OB(3)

D30     =  30,    // D30

D31     =  31,    // D31

D32     =  32,    // D32

D33     =  33,    // D33

D34     =  34,    // D34

D35     =  35,    // D35

D36     =  36,    // D36

D37     =  37,    // D37

//IB(4) or OB(4)

D38     =  38,    // D38

D39     =  39,    // D39

D40     =  40,    // D40

D41     =  41,    // D41

D42     =  42,    // D42

D43     =  43,    // D43

D44     =  44,    // D44

D45     =  45,    // D45

//IB(5) or OB(5)

D46     =  46,    // D46

D47     =  47,    // D47

D48     =  48,    // D48

D49     =  49,    // D49

D50     =  50,    // D50

D51     =  51,    // D51

D52     =  52,    // D52

D53     =  53,    // D53

//IB(6) or OB(6)

// Addr  0x20 A

//IB(7) or OB(7)

//  Addr  0x20 B

//IB(8) or OB(8)

//  Addr  0x21 A

//IB(9) or OB(9)

//  Addr  0x21 B

//IB(10) or OB(10)

//  Addr  0x22 A

//IB(11) or OB(11)

//  Addr  0x22 B

//IB(12) or OB(12)

//  Addr  0x23 A

//IB(13) or OB(13)

//  Addr  0x23 B

 

//IB(14) or OB(14)

//  Addr  0x24 A

//IB(15) or OB(15)

//  Addr  0x24 B

//IB(16) or OB(16)

//  Addr  0x25 A

//IB(17) or OB(17)

//  Addr  0x25 B

//IB(18) or OB(18)

//  Addr  0x26 A

//IB(19) or OB(19)

//  Addr  0x26 B

//IB(20) or OB(20)

//  Addr  0x27 A

//IB(21) or OB(21)

//  Addr  0x27 B

 

 

 

 

 

 

 

 

// Analog Pins

//------------

ADC0     = A0,  // A0

ADC1     = A1,  // A1

ADC2     = A2,  // A2

ADC3     = A3,  // A3

ADC4     = A4,  // A4

ADC5     = A5,  // A5

 

ADC8     = A8,  // A8

ADC9     = A9,  // A9

ADC10     = A10,  // A10

ADC11     = A11,  // A11

ADC12     = A12,  // A12

ADC13     = A13,  // A13

ADC14     = A14,  // A14

ADC15     = A15;  // A15

// Address the bug in the Arduino IDE which does not handle re-defined analog pins.

// Redefined analog pin names do not work in digitalWrite calls and for loop indexes.

// This has been confirmed by extensive testing.

// D12 and D13 are included in the table for ease of coding bit extraction of the high byte

//-----------------------------------------------------------------------------------------

const byte  D12i    = 0,  // Index to actual internal pin value to keep the compiler happy

            D13i    = 1,

            ADC0i     = 2,

            ADC1i     = 3,

            ADC2i     = 4,

            ADC3i     = 5,

            ADC4i     = 6,

            ADC5i     = 7,

 

            ADC8i    = 0,  // Index to actual internal pin value to keep the compiler happy

            ADC9i    = 1,

            ADC10i     = 2,

            ADC11i     = 3,

            ADC12i     = 4,

            ADC13i     = 5,

            ADC14i     = 6,

            ADC15i     = 7;

 

 

// Pin name values which work for analog pins.   PDxx included for coding convenience

//-----------------------------------------------------------------------------------

const byte numAPorts = 16;

int APortMap[numAPorts] = {D12, D13, A0, A1, A2, A3, A4, A5, A8, A9, A10, A11, A12, A13, A14, A15};

 

boolean debugMode  = true,

        debugMode1 = false;

 

// Protocol characters

//--------------------

const char STX    = 0x02,

           ETX    = 0x03,

           DLE    = 0x10,

           SYN    = 0xFF,

 

           // Read responses

           //---------------

           respNone    = 0,  //  No character received

           respErr     = 1,  //  Error occured

           respIgnore  = 2,  //  Not for this node, flush to ETX

           respInit    = 3,  //  "I" Message

           respPoll    = 4,  //  "P" Message

           respRead    = 5,  //  "R" Message

           respTransmit = 6; //  "T" Message

 

// Debug monitor speed

//--------------------

const int MON_SPEED      = 9600;

 

const int DEBOUNCE_DELAY = 2;

 

// Timer variables

//----------------

long previousMS = 0;

unsigned long currentMS = 0;

 

//---------------

// Node variables

//---------------

const char cpNODE_NDP = 'C';    // Node Definition Parameter for a cpNode - Control Point Node

 

byte  UA,           // Node address, stored as UA + UA_Offset

      matchID;      // Incomming address in message to match to node UA

int   c;            // Handy character variable

 

const int UA_Offset = 65,       // 0x41 Per CMRI protocol spec

          CMRInet_BufSize  = 260;  //  Max SUSIC + 4 pad

 

// Data buffers

//-------------

const byte cpNode_bufSize = 20,  //  cpNode (6) + IOXx8 (16) + 4 pad

           IO_bufsize    = 20,   //  cpNode (6) + IOXx8 (16) + 4 pad

#ifdef BASE_NODE

           maxIB         = 2,    //  Max number of onboard input bytes

           maxOB         = 2;    //  Max number of onboard output bytes

#endif

 

#ifdef BASE_NODE_8IN8OUT

maxIB         = 2,    //  Max number of onboard input bytes

maxOB         = 2;    //  Max number of onboard output bytes

#endif

 

#ifdef BASE_NODE_8OUT8IN   // cpNode System Test Configuration

maxIB         = 2,    //  Max number of onboard input bytes

maxOB         = 2;    //  Max number of onboard output bytes

#endif

 

#ifdef BASE_NODE_12OUT4IN

maxIB         = 2,    //  Max number of onboard input bytes

maxOB         = 2;    //  Max number of onboard output bytes

#endif

 

#ifdef BASE_NODE_16IN

maxIB         = 2,    //  Max number of onboard input bytes

maxOB         = 2;    //  Max number of onboard output bytes

#endif

 

#ifdef BASE_NODE_16OUT

maxIB         = 2,    //  Max number of onboard input bytes

maxOB         = 2;    //  Max number of onboard output bytes

#endif

 

#ifdef BASE_NODE_RSMC

maxIB         = 2,    //  Max number of onboard input bytes

maxOB         = 2;    //  Max number of onboard output bytes

#endif

 

#ifdef BASE_NODE_RSMC_LOCK

maxIB         = 2,    //  Max number of onboard input bytes

maxOB         = 2;    //  Max number of onboard output bytes

#endif

 

 

 

#ifdef BASE_NODE_16IN40OUT

maxIB         = 2,    //  Max number of onboard input bytes

maxOB         = 5;    //  Max number of onboard output bytes

 

#endif

 

 

byte       CMRInet_Buf[CMRInet_BufSize],   // CMRI message buffer

 

           OB[IO_bufsize],  // Output bits  HOST to NODE

           IB[IO_bufsize];  // Input bits   NODE to HOST

 

byte       nIB,             // Total configured input bytes

           nOB;             // Total configured output bytes

 

unsigned long DL  = 0;      // Transmit character delay value in 10 us increments

int           DLH = 0,

              DLL = 0;

 

/**********************************************************************/

/**********                      IOX VARIABLES               **********/

/**********************************************************************/

#ifdef USE_IOX

 

// If using the I/O expander, include the i2c library

//---------------------------------------------------

#include <Wire.h>

 

const byte max_IOX = 16,  // device = MCP28017 chip

 

           // I2C command registers

           //----------------------

           ACTIVELOW_REG_BASE  = 0x02,  // Port A, Port B = 0x02

           PULLUP_REG_BASE     = 0x0C,  // Port A, Port B = 0x0D

           GPIO_REG_BASE       = 0x12,  // Port A, Port B = 0x13

 

           SET_PORT_OUTPUT     = 0x00,

           SET_PORT_INPUT      = 0xFF,

           SET_PORT_PULLUPS    = 0xFF,

           SET_PORT_ACTIVELOW  = 0xFF;

 

// Board Addresses (duplicated for ease of indexing)

const int IOX_devAddr[max_IOX] = {0x20, 0x20, 0x21, 0x21, 0x22, 0x22, 0x23, 0x23,

                                  0x24, 0x24, 0x25, 0x25, 0x26, 0x26, 0x27, 0x27

                                 },

 

                                 // I2C port direction bytes 0 = OUTPUT  1 = INPUT  -1 = Not Assigned

                                 //------------------------------------------------------------------

                                 // Addr  0x20     0x21     0x22     0x23     0x24     0x25     0x26     0x27

                                 // Port  A  B     A  B     A  B     A  B     A  B     A  B     A  B     A  B

                                 IOX_ioMap[max_IOX]   =   {  0, 0,     0, 0,   0, 0,    0, 0,    0, 0,    0, 0,    0, 0,    0, 0 };

 

// Data byte mapping table created from devAddr/ioMap configuration

// Created by Configure_IOX_Ports

 

// [port index][0] is I2C chip address

// [port index][1] is I2C port 0=A, 1=B

//-----------------------------------------------------------------

const byte I2C_ADDR = 0,

           I2C_PORT = 1;

 

int IOX_Output_Map[max_IOX][2],

    IOX_Input_Map[max_IOX][2];

 

// Actual count of defined I2C input and output ports

//---------------------------------------------------

byte numIOX_OUT = 0,

     numIOX_IN = 0;

 

// I2C input and output buffers

//-----------------------------

int IOX_inBuf[max_IOX + 2],

    IOX_outBuf[max_IOX + 2];

 

 

/*************************************************************************

***********                IOX SUPPORT ROUTINES                 **********

**************************************************************************

  For each defined IOX board, setup the two ports for direction.

  Input ports will have the weak pull up enabled.

 

  Packing and Unpacking routines will place data bytes sequentially in

  the IOX buffers in the order of device address.

**************************************************************************/

 

//------------------------------------------------------------------------

// Collect the input/output byte assignments for the number of IOX's

// Build the tables used for processing the data bytes

//------------------------------------------------------------------------

void Configure_IOX_Ports()

{

  byte i = 0,

       j, k;

 

  // Count the number of defined ports by direction.

  // Setup the input and output IOX byte map for the CMRI buffers

  //-------------------------------------------------------------

  while (i < max_IOX)

  {

    for (j = 0; j < 2; j++)

    {

      k = i + j;

      switch (IOX_ioMap[k])

      {

        case  0:   // Outputs

          IOX_Output_Map[numIOX_OUT][I2C_ADDR] = IOX_devAddr[k];

          IOX_Output_Map[numIOX_OUT++][I2C_PORT] = j;

          break;

        case  1:   // Inputs

          IOX_Input_Map[numIOX_IN][I2C_ADDR] = IOX_devAddr[k];

          IOX_Input_Map[numIOX_IN++][I2C_PORT] = j;

          break;

        default: ; // No Port in slot

      }

    }

    i = i + 2;

  }

 

}  // CONFIGURE IOX PORTS

 

 

//-------------------------------------------------------------------------------

// For each of the defined OUTPUT bytes, set the register parameters for the chip

//-------------------------------------------------------------------------------

void Setup_IOX_Outputs()

{

  byte i;

 

  for (i = 0; i < numIOX_OUT; i++)

  {

    Wire.beginTransmission(IOX_Output_Map[i][I2C_ADDR]);  // Board Address

    Wire.write(IOX_Output_Map[i][I2C_PORT]);              // A or B Port

    Wire.write(SET_PORT_OUTPUT);                          // Command to set to Output

    Wire.endTransmission();

  }

 

}  // SET UP IOX OUTPUTS

 

 

//--------------------------------------------------------------------------

// Unpack the bytes sent in the transmit message and output to the I2C ports

//--------------------------------------------------------------------------

void Unpack_IOX_Outputs()

{

  byte i, j, n;

 

  // Set up values to assign CMRI output bytes to I2C bytes

  //-------------------------------------------------------

  if (numIOX_OUT > 0)

  {

    i = nOB;      // Last onboard output byte index

    j = 0;        // Index into stored I2C output byte

    n = 0;        // Byte transfer count

 

    // Transfer CMRI output bytes to the I2C buffer

    //---------------------------------------------

    while (n < numIOX_OUT)

    {

      IOX_outBuf[j++] = CMRInet_Buf[i++];

      n++;

    }

 

    // Write the bytes to the I2C chips

    //---------------------------------

    for (i = 0; i < numIOX_OUT; i++)

    {

      Wire.beginTransmission(IOX_Output_Map[i][I2C_ADDR]);        // Board Address

      Wire.write(GPIO_REG_BASE + IOX_Output_Map[i][I2C_PORT]);    // A or B Port

      Wire.write(IOX_outBuf[i]);                                  // Data to send

      Wire.endTransmission();

    }

  }

 

}  // UNPACK IOX OUTPUTS

 

 

//-------------------------------------------------------------------------------

// For each of the defined INPUT bytes, set the register parameters for the chip

//-------------------------------------------------------------------------------

void Setup_IOX_Inputs()

{

  byte i,

       addr, port;

 

 

  for (i = 0; i < numIOX_IN; i++)

  {

    addr =  IOX_Input_Map[i][I2C_ADDR];

    port =  IOX_Input_Map[i][I2C_PORT];

 

    // Set port to Inputs

    //-------------------

    Wire.beginTransmission(addr);

    Wire.write(port);

    Wire.write(SET_PORT_INPUT);

    Wire.endTransmission();

 

    // Set port to use weak pullups

    //-----------------------------

    Wire.beginTransmission(addr);

    Wire.write(PULLUP_REG_BASE + port);

    Wire.write(SET_PORT_PULLUPS);

    Wire.endTransmission();

 

    // Set port to active LOW polarity

    //--------------------------------

    Wire.beginTransmission(addr);

    Wire.write(ACTIVELOW_REG_BASE + port);

    Wire.write(SET_PORT_ACTIVELOW);

    Wire.endTransmission();

  }

 

}  // SET UP IOX INPUTS

 

 

//-----------------------------------------------------------------------------------

// Read the input bytes from the I2C ports and collect them into the I2C input buffer

// The bits will be latched and cleared when sent in the poll response.

//-----------------------------------------------------------------------------------

void Pack_IOX_Inputs()

{

  byte i;

 

  // Transfer the input bytes to the I2C buffer

  //-------------------------------------------

  if (numIOX_IN > 0)

  {

    for (i = 0; i < numIOX_IN; i++)

    {

      Wire.beginTransmission(IOX_Input_Map[i][I2C_ADDR]);        // Board Address

      Wire.write(GPIO_REG_BASE + IOX_Input_Map[i][I2C_PORT]);    // A or B Port

      Wire.endTransmission();

 

      Wire.requestFrom(IOX_Input_Map[i][I2C_ADDR], 1);           // Data to read

      IOX_inBuf[i] = IOX_inBuf[i] | Wire.read();                 // Latch the inputs

      delay(DEBOUNCE_DELAY);

    }

  }

 

}  // PACK IOX INPUTS

 

#endif

 

 

 

 

/*************************************************************************/

/**********            GENERAL SUPPORT ROUTINES                 **********/

/*************************************************************************/

//-----------------------

//  Get the available ram

//-----------------------

int freeRam()

{

  extern int __heap_start, *__brkval;

  int v;

  return (int) &v - (__brkval == 0 ? (int) &__heap_start : (int) __brkval);

}  // FREE RAM

 

 

/*---------------------------------------------------------------------

  Perform the base cpNode setup at bootup

  -----------------------------------------------------------------------*/

void SetUp_Node()

{

 

  // Set the number of onboard I/O bytes

  //-------------------------------------

  nIB = maxIB;

  nOB = maxOB;

 

}  // SET UP NODE

 

 

/*----------------------------------------------

  Set the Node ID

  ------------------------------------------------*/

void SetNodeAddr(byte nodeAddr)

{

 

  // Set node address to 127 if an invalid decimal address is passed

  //----------------------------------------------------------------

  if ((nodeAddr < 0) || (nodeAddr > 127))

    nodeAddr = 127;

 

  UA = nodeAddr + UA_Offset;

 

}  // SET NODE ADDR

 

 

/*************************************************************************/

/**********              DATA BIT PACK/UNPACK ROUTINES          **********/

/*************************************************************************/

 

/*************************************************************************/

/******       cpNode onboard I/O bit mapping for 16 bits            ******/

/******       Only one ioMap is active per node configuration       ******/

/*************************************************************************/

 

/*-------------------------------------------------------------------------

  The input routines collect the bits from the digitalRead() calls and

  puts the them into the correct bytes for transmission.

 

  The ports are read twice with a delay between for input debounce.

 

  Two bytes are always stored.

  ---------------------------------------------------------------------------*/

void Pack_Node_Inputs()

{

  byte i, k, bits,

       n;

 

  for (i = 0; i < 2; i++)

  {

#ifdef BASE_NODE

    // IB0 6 bits     IB1 All Zero

    // 00IIIIII       00000000

    //----------------------------

    bits = 0;

    for (n = 0, k = ADC5i; n < 6; n++)

      bits = bits | (!digitalRead(APortMap[k--]) << (5 - n));

    IB[0] = bits;

    IB[1] = 0;

#endif

 

#ifdef BASE_NODE_8IN8OUT

    // IB0 8 bits     IB1 All Zero

    // IIIIIIII       00000000

    //----------------------------

    bits = 0;

    for (n = 0, k = D11; n < 8; n++)

      bits = bits | (!digitalRead(k--) << (7 - n));

    IB[0] = bits;

    IB[1] = 0;

#endif

 

#ifdef BASE_NODE_8OUT8IN   // cpNode System Test Configuration

    // D12 - A5 are inputs

    // IB0 8 bits     IB1 All Zero

    // IIIIIIII       00000000

    //----------------------------

    bits = 0;

    for (n = 0, k = ADC5i; n < 8; n++)

      bits = bits | (!digitalRead(APortMap[k--]) << (7 - n));

    IB[0] = bits;

    IB[1] = 0;

#endif

 

#ifdef BASE_NODE_12OUT4IN

    // IB0 4 bits     IB1 All Zero

    // 0000IIII       00000000

    //----------------------------

    bits = 0;

    for (n = 0, k = ADC5i; n < 4; n++)

      bits = bits | (!digitalRead(APortMap[k--]) << (3 - n));

    IB[0] = bits;

    IB[1] = 0;

#endif

 

#ifdef BASE_NODE_16IN

    // IB0 8 bits     IB1 8 bits

    // IIIIIIII       IIIIIIII

    //----------------------------

    bits = 0;

    for (n = 0, k = D11; n < 8; n++)

      bits = bits | (!digitalRead(k--) << (7 - n));

    IB[0] = bits;

 

    bits = 0;

    for (n = 0, k = ADC5i; n < 8; n++)

      bits = bits | (!digitalRead(APortMap[k--]) << (7 - n));

    IB[1] = bits;

#endif

 

#ifdef BASE_NODE_16OUT

    // IB0 All Zero   IB1 All Zero

    // 00000000       00000000

    //----------------------------

    IB[0] = 0;

    IB[1] = 0;

#endif

 

#ifdef BASE_NODE_RSMC

    // IB0 5 bits     IB1 All Zero

    // 000IIIII       00000000

    //----------------------------

    bits = 0;

    for (n = 0, k = ADC5i; n < 5; n++)

      bits = bits | (!digitalRead(APortMap[k--]) << (4 - n));

    IB[0] = bits;

    IB[1] = 0;

#endif

 

#ifdef BASE_NODE_RSMC_LOCK

    // IB0 4 bits     IB1 All Zero

    // 0000IIII       00000000

    //----------------------------

    bits = 0;

    for (n = 0, k = ADC5i; n < 4; n++)

      bits = bits | (!digitalRead(APortMap[k--]) << (3 - n));

    IB[0] = bits;

    IB[1] = 0;

#endif

 

 

#ifdef BASE_NODE_16IN40OUT

    // IB0 8 bits     IB1 8 bits

    // IIIIIIII       IIIIIIII

    //----------------------------

    bits = 0;

    for (n = 0, k = D11; n < 8; n++)

      bits = bits | (!digitalRead(k--) << (7 - n));

    IB[0] = bits;

 

    bits = 0;

    for (n = 0, k = ADC5i; n < 8; n++)

      bits = bits | (!digitalRead(APortMap[k--]) << (7 - n));

    IB[1] = bits;

 

    IB[2] = 0;

    IB[3] = 0;

    IB[4] = 0;

    IB[5] = 0;

    IB[6] = 0;

#endif

    // Wait for debounce time

    //-----------------------

    delay(DEBOUNCE_DELAY);

  }

 

}  // PACK NODE INPUTS

 

 

/*-------------------------------------------------------------------------

  Unpack the first byte of output bit stream.  Called first by most of the

  unpacking modes.

  ---------------------------------------------------------------------------*/

void Unpack_Lower8()

{

  byte n, j, k,

       bits;

 

  bits = OB[0];

  for (j = 0, k = D4; k < D12; k++)

  {

    n = ((bits >> j++) &  0x01);

    digitalWrite(k, n);

  };

 

}  // UNPACK LOWER 8

 

 

/*-------------------------------------------------------------------------

  The output routines takes received bits from the ouput buffer and

  writes them to the correct output port using digitalWrite().

 

  The correct formatting routine is based upon the value of cpNode_ioMap

  ---------------------------------------------------------------------------*/

void Unpack_Node_Outputs()

{

  byte i, j, k,

       n, bits;

 

  // Move the received bytes to the output buffer

  //---------------------------------------------

  for (i = 0; i < maxOB; i++)

    OB[i] = CMRInet_Buf[i];

 

#ifdef BASE_NODE

  // OB1 8 bits     OB2 2 bits      -> NODE

  // OOOOOOOO       xxxxxxOO

  //----------------------------

  Unpack_Lower8();

  bits = OB[1];

  for (j = 0, k = D12i; k < ADC0i; k++) // Loop limit one bit before last IO pin

  { // stops on D13

    n = ((bits >> j++) &  0x01);

    digitalWrite(APortMap[k], n);

  };

#endif

 

#ifdef BASE_NODE_8IN8OUT

  // OB1 8 bits     OB2 All Zero    -> NODE

  // OOOOOOOO       xxxxxxxx

  //----------------------------

  bits = OB[0];

  for (j = 0, k = D12i; k <= ADC5i; k++) // v1.2 output bits to D12 to A5

  { // v1.3 loop to include high order bit

    n = ((bits >> j++) &  0x01);

    digitalWrite(APortMap[k], n);

  };

#endif

 

#ifdef BASE_NODE_8OUT8IN  // cpNode System Test Configuration

  // D4 - D11 are ouptuts

  // OB1 8 bits     OB2 All Zero

  // OOOOOOOO       xxxxxxxx

  //----------------------------

  Unpack_Lower8();

#endif

 

#ifdef BASE_NODE_12OUT4IN

  // OB1 8 bits     OB2 4 bits      -> NODE

  // OOOOOOOO       xxxxOOOO

  //----------------------------

  Unpack_Lower8();

  bits = OB[1];

  for (j = 0, k = D12i; k < ADC2i; k++) // Loop limit one bit before last IO pin  TVerberg

  { // stops on A1

    n = ((bits >> j++) &  0x01);

    digitalWrite(APortMap[k], n);

  };

#endif

 

#ifdef BASE_NODE_16IN

  // OB1 All Zero   OB2 All Zero    -> NODE

  // xxxxxxxx       xxxxxxxx

  //----------------------------

#endif

 

#ifdef BASE_NODE_16OUT

  // OB1 8 bits     OB2 8 bits      -> NODE

  // OOOOOOOO       OOOOOOOO

  //----------------------------

  Unpack_Lower8();

  bits = OB[1];

  for (j = 0, k = D12i; k <= ADC5i; k++) // v1.3 loop to include high order bit

  {

    n = ((bits >> j++) &  0x01);

    digitalWrite(APortMap[k], n);

  };

#endif

 

#ifdef BASE_NODE_RSMC

  // OB1 8 bits       OB2 3 bits      -> NODE

  // OOOOOOOO         xxxxxOOO

  //----------------------------

  Unpack_Lower8();

  bits = OB[1];

  for (j = 0, k = D12i; k < PA1i; k++)

  {

    n = ((bits >> j++) &  0x01);

    digitalWrite(APortMap[k], n);

  };

#endif

 

#ifdef BASE_NODE_RSMC_LOCK

  // OB1 8 bits       OB2 4 bits      -> NODE

  // OOOOOOOO         xxxxOOOO

  //----------------------------

  Unpack_Lower8();

  bits = OB[1];

  for (j = 0, k = D12i; k < ADC2i; k++)

  {

    n = ((bits >> j++) &  0x01);

    digitalWrite(APortMap[k], n);

  };

#endif

 

 

 

#ifdef BASE_NODE_16IN40OUT

 

  // OB1 8 bits     OB2 8 bits      -> NODE

  // OOOOOOOO       OOOOOOOO

  //----------------------------

  {

    byte n, j, k,

         bits;

    bits = OB[0];

    for (j = 0, k = A8; k <= A15; k++)

    {

      n = ((bits >> j++) &  0x01);

      digitalWrite(k, n);

    };

    bits = OB[1];

    for (j = 0, k = D22; k <= D29; k++)

    {

      n = ((bits >> j++) &  0x01);

      digitalWrite(k, n);

    };

    bits = OB[2];

    for (j = 0, k = D30; k <= D37; k++)

    {

      n = ((bits >> j++) &  0x01);

      digitalWrite(k, n);

    };

 

    bits = OB[3];

    for (j = 0, k = D38; k <= D45; k++)

    {

      n = ((bits >> j++) &  0x01);

      digitalWrite(k, n);

    };

 

    bits = OB[4];

    for (j = 0, k = D46; k <= D53; k++)

    {

      n = ((bits >> j++) &  0x01);

      digitalWrite(k, n);

    };

 

  }

 

 

 

#endif

 

 

 

}  // UNPACK NODE OUTPUTS

 

 

/*************************************************************************/

/**********            CMRInet SERIAL SUPPORT ROUTINES          **********/

/*************************************************************************/

 

//-----------------------------------

//CMRInet cpNode Option Bit Processor

//-----------------------------------

void Process_cpNode_Options()

{

}  // PROCESS CPNODE OPTIONS

 

 

/*-----------------------------------------------------------------------------------------

  Perform any initialization and setup using the initialization message

  NDP must be a "C" (cpNode) for initialization to be done

 

 - cpNode Initialization Message (I)

  SYN SYN STX <UA> <I><NDP> <DLH><DLL> <opts1><opts2> <NIN><NOUT> <000000><ETX>

  -------------------------------------------------------------------------------------------*/

void Initialize_cpNode()

{

 

  // Set up transmit delay

  // 1 unit of delay(DL) is 10 microseconds

  //----------------------------------------

  DLH = CMRInet_Buf[1];

  DLL = CMRInet_Buf[2];

  DL  = (DLH * 256) + DLL;    // Transmit character delay value in 10 us increments

  DL  = DL * 10;

 

  /*

    if (debugMode1)

     {

      Serial.print("DLH=");    Serial.print(DLH);

      Serial.print(" DLL=");   Serial.print(DLL);

      Serial.print(" DL/10="); Serial.print(DL/10);

      Serial.print(" DL=");    Serial.print(DL);

     }

  */

 

  // Check if initialize message is for a cpNode

  // if so, process any options

  //--------------------------------------------

  if (cpNODE_NDP == CMRInet_Buf[0])

  {

    /*

           if (debugMode1)

           {

            Serial.print("  option_Byte0 = "); Serial.print(option_byte0,HEX);

            Serial.print("  option_Byte1 = "); Serial.println(option_byte1,HEX);

           }

    */

    Process_cpNode_Options();

  }

 

}  // INITIALIZE CP NODE

 

 

/*-----------------------------------------------------

  This function flushes the serial read until an ETX

  is seen.  This is used to ignore any inbound messages

  not addressed to the node or to re-sync the protocol

  parser for a garbled message.

  -------------------------------------------------------

  void Flush_CMRI_To_ETX()

  {

  do

  {

   if (Serial1.available())

    c = Serial1.read();

  } while (c != ETX);

 

  }  // FLUSH CMRI TO ETX

*/

 

/*-----------------------------------------------------

  This function flushes the serial read until an ETX

  is seen or the input serial buffer is empty.

  This is used to ignore any inbound messages

  not addressed to the node or to re-sync the protocol

  parser for a garbled message.

  -------------------------------------------------------*/

void Flush_CMRInet_To_ETX()

{

  boolean done = false;

 

  while (!done)

  {

    if (Serial1.available())

    {

      if (Serial1.read() == ETX) done = true;

    }

    else

      done = true;

  }

 

}  // FLUSH CMRInet TO ETX

 

 

/*--------------------------------------------------------

  Read a byte from the specified serial port.

  Return the character read

  ----------------------------------------------------------*/

char Read_CMRI_Byte()

{

  while (true)

  {

    if (Serial1.available() > 0)

      return char(Serial1.read());

  };

}  // READ CMRI BYTE

 

 

/*----------------------------------------------------------------------

  Read the message from the Host and determine if the message is

  for this node.  The whole message is read, any data to be processed

  is stored in CMRInet_Buf[], stripped of protocol characters.

 

  If the node address does not match, the data is ignored.

 

  The data message body is processed by an appropriate message handler.

 

  - Initialization Packet (I)

  SYN SYN STX <UA> <I> <NDP> <dH><dL> <NS> <CT(1)><CT(NS)> ETX

 

  - Poll for Data (P)

  SYN SYN STX <UA> <P> ETX

 

  - Read Data (R)

  SYN SYN STX <UA> <R> <IB(1)><IB(NS)> ETX

  ----------------------------------------------------------------------*/

int CMRI_Read()

{

  byte i,

       resp;

  int  inCnt;

 

  boolean reading = true,

          inData = false;

 

  //-----------------------------------

  // Check input buffer for a character

  //-----------------------------------

  if (Serial1.available() <= 0)

    return respNone;

 

  //--------------------

  // Process the message

  //--------------------

  matchID = 0;

  inCnt = 0;

 

  do

  {

    c = Read_CMRI_Byte();  // read the byte

    switch ( int(c) )

    {

      case STX:         // Start of message header, start parsing protocol message

        //        if (debugMode) Serial.print("STX ");

 

        // Read node address and message type

        //-----------------------------------

        matchID = Read_CMRI_Byte();  // Node Address

        //        if (debugMode) {Serial.print(matchID-UA_Offset); Serial.print(" ");}

 

        // If node ID does not match, exit and flush to ETX in outer loop

        //---------------------------------------------------------------

        if (matchID != UA)

        {

          //          if (debugMode) {Serial.print("No Match "); Serial.println(matchID); }

          resp = respIgnore;

          reading = false;

        }

        else

        {

          // Set response code based upon message type

          //------------------------------------------

          c = Read_CMRI_Byte();        // Message Type

          //       if (debugMode) { Serial.print(char(c)); Serial.print(" ");}

 

          switch ( c )

          {

            case 'I':          // Initialization

              resp = respInit;

              break;

            case 'P':          // Poll

              resp = respPoll;

              break;

            case 'R':          // Read

              resp = respRead;

              break;

            case 'T':          // Write (Transmit)

              resp = respTransmit;

              //           if (debugMode) { Serial.print(":"); i=Serial1.available(); Serial.print(i,HEX); }

              break;

            default:           // Unknown - Error

              resp = respErr;

              reading = false;

              break;

          }

 

          // Completed the header, go into message data mode

          //------------------------------------------------

          inData = true;

        }

        break;

 

      case ETX:         // End of message, read complete

        //        if (debugMode) Serial.print(" ETX ");

        reading = false;

        break;

 

      case DLE:         // Read the next byte regardless of value and store it

        //        if (debugMode) Serial.print("DLE ");

        CMRInet_Buf[inCnt++] = Read_CMRI_Byte();

        //        if (debugMode) {Serial.print(byte(CMRInet_Buf[inCnt-1]),HEX); Serial.print(" ");}

        break;

 

      case SYN:         // Sync character Ignore it if not reading data

        if (inData) CMRInet_Buf[inCnt++] = c;

        //        if (debugMode) Serial.print("SYN ");

        break;

 

      default:  // Stuff the data character into the receive buffer

        //       if (debugMode) {Serial.print("{"); Serial.print(byte(c),HEX); Serial.print("}");}

        CMRInet_Buf[inCnt++] = c;

        break;

    }

 

    // Check for buffer overrun and terminate the read if true

    //--------------------------------------------------------

    if (inCnt > CMRInet_BufSize)

    {

      reading = false;

      resp = respErr;

      //        if (debugMode) { Serial.print("\nBuffer Overrun inCnt = "); Serial.println(inCnt); };

   }

 

  } while (reading);

 

  // Null terminate the input buffer

  //--------------------------------

  CMRInet_Buf[inCnt] = 0;

 

  //---------------------------------------------------------

  // Match the node address in the message to the UA+65 value

  // if no match, ignore message, not addessed to this node

  //---------------------------------------------------------

  //  if (debugMode) { Serial.print("\n ->inCnt = "); Serial.println(inCnt); };

 

  return resp;

 

}  // CMRI READ

 

 

/*----------------------------------------------------------

  This function sends the input bytes to the host in response

  to a poll message.  Bytes are moved from the IB(n) buffer

  to the transmit buffer.

  DLE characters are inserted for data values which are

  protocol characters.

 

  - Read Data (R) Message

  SYN SYN STX <UA> <R><IB(1)><IB(NS)> ETX

 

  NB: Change to int if longer buffers are implemented

  ------------------------------------------------------------*/

void CMRI_Poll_Resp()

{

  byte i = 0,

       j = 0;

 

  // Packet Header

  //--------------

  CMRInet_Buf[i++] = SYN;

  CMRInet_Buf[i++] = SYN;

  CMRInet_Buf[i++] = STX;

 

  // Message Header

  //---------------

  CMRInet_Buf[i++] = UA;

  CMRInet_Buf[i++] = 'R';

 

  // Load the onboard input bytes into the output buffer

  //----------------------------------------------------

  if (nIB > 0)

  {

    for (j = 0; j < nIB; j++)

    {

      c = IB[j];  // Insert a DLE if the output byte value is a protocol character

      switch (c)

      {

        case STX:

        case ETX:

        case DLE:

          //       case SYN:   Syncs are ignored to conform to the published protocol

          CMRInet_Buf[i++] = DLE;

          break;

      }

      CMRInet_Buf[i++] = c;

      IB[j] = 0;   // Clear the latched inputs

    }

  }

 

#ifdef USE_IOX

  // Load the IOX input bytes into the output buffer

  //------------------------------------------------

  if (numIOX_IN > 0)

  {

    for (j = 0; j < numIOX_IN; j++)

    {

      c = IOX_inBuf[j];  // Insert a DLE if the output byte value is a protocol character

      switch (c)

      {

        case STX:

        case ETX:

        case DLE:

          //       case SYN:    Syncs are ignored to conform to the published protocol

          CMRInet_Buf[i++] = DLE;

          break;

      }

      CMRInet_Buf[i++] = c;

      IOX_inBuf[j] = 0;   // Clear the latched inputs

    }

  }

#endif

 

  // Add the ETX and send the complete buffer

  //-----------------------------------------

  CMRInet_Buf[i++] = ETX;

 

  // Send the packet to the host

  //----------------------------

  for (j = 0; j < i; j++)

  {

    Serial1.write(CMRInet_Buf[j]);

 

    // If a transmit delay was set, delay microseconds

    //------------------------------------------------

    if (DL > 0)

      delayMicroseconds( DL );   // value in microseconds

  }

  /*

    if (debugMode)

     {

       Serial.print("Poll Response "); Serial.print("nIDB= "); Serial.print(nIB,HEX); Serial.print(" ");

       for (j=0; j<i; j++) {Serial.print(CMRInet_Buf[j],HEX); Serial.print(" "); }

       Serial.println();

     }

  */

}  // CMRI POLL RESP

 

 

/************************************************

*******            Setup Code          **********

*************************************************/

void setup()

{

 

 

 

  byte i, j;

  //

 

 

 

 

#ifdef DCC_Accessory_Decoder

 

  Dcc.init( MAN_ID_DIY, 601, FLAGS_OUTPUT_ADDRESS_MODE | FLAGS_DCC_ACCESSORY_DECODER, 00);

 

#endif

 

 

  // RS-485 Port

  //------------

  pinMode(RX, INPUT);          // D0  CMRI RS485 Receive

  pinMode(TX, OUTPUT);         // D1  CMRI RS485 Transmit

 

  // I2C Ports, actual setup done by the Wire library

  //-------------------------------------------------

  pinMode(i2C_SDA, OUTPUT);    // D2  SDA (data)

  pinMode(i2c_SCL, OUTPUT);    // D3  SCL (clock)

 

#ifdef BASE_NODE    //----------  Base Node assignments ----------

  pinMode(D4,  OUTPUT);       // D4  SG1A-R

  pinMode(D5,  OUTPUT);       // D5  SG1A-Y

  pinMode(D6,  OUTPUT);       // D6  SG1A-G

  pinMode(D7,  OUTPUT);       // D7  SG1B-R

  pinMode(D8,  OUTPUT);       // D8  SG1B-Y

  pinMode(D9,  OUTPUT);       // D9  SG2-R

  pinMode(D10, OUTPUT);       // D10 SG2-Y

  pinMode(D11, OUTPUT);       // D11 SG2-G

  pinMode(D12, OUTPUT);       // D12 SG3-R

  pinMode(D13, OUTPUT);       // D13 SG3-Y

 

  pinMode(A0,  INPUT_PULLUP);  // A0  TC1  Track Circuit 1

  pinMode(A1,  INPUT_PULLUP);  // A1  TC2  Track Circuit 2

  pinMode(A2,  INPUT_PULLUP);  // A2  OS1  OS Circuit 1

  pinMode(A3,  INPUT_PULLUP);  // A3  SW1  Turnout throw input

  pinMode(A4,  INPUT_PULLUP);  // A4  AUX1

  pinMode(A5,  INPUT_PULLUP);  // A5  AUX2

#endif

 

#ifdef BASE_NODE_8IN8OUT //----------  Base Node 8 IN 8 OUT ----------

  pinMode(D4,  INPUT_PULLUP);        // D4

  pinMode(D5,  INPUT_PULLUP);        // D5

  pinMode(D6,  INPUT_PULLUP);        // D6

  pinMode(D7,  INPUT_PULLUP);        // D7

  pinMode(D8,  INPUT_PULLUP);        // D8

  pinMode(D9,  INPUT_PULLUP);        // D9

  pinMode(D10, INPUT_PULLUP);        // D10

  pinMode(D11, INPUT_PULLUP);        // D11

 

  pinMode(D12, OUTPUT);              // D12

  pinMode(D13, OUTPUT);              // D13

  pinMode(A0,   OUTPUT);              // A0

  pinMode(A1,   OUTPUT);              // A1

  pinMode(A2,   OUTPUT);              // A2

  pinMode(A3,   OUTPUT);              // A3

  pinMode(A4,   OUTPUT);              // A4

  pinMode(A5,   OUTPUT);              // A5

#endif

 

#ifdef BASE_NODE_8OUT8IN //----------  Base Node 8 OUT 8 IN ----------

  pinMode(D4,  OUTPUT);              // D4

  pinMode(D5,  OUTPUT);              // D5

  pinMode(D6,  OUTPUT);              // D6

  pinMode(D7,  OUTPUT);              // D7

  pinMode(D8,  OUTPUT);              // D8

  pinMode(D9,  OUTPUT);              // D9

  pinMode(D10, OUTPUT);              // D10

  pinMode(D11, OUTPUT);              // D11

 

  pinMode(D12, INPUT_PULLUP);        // D12

  pinMode(D13, INPUT_PULLUP);        // D13

  pinMode(A0,   INPUT_PULLUP);        // A0

  pinMode(A1,   INPUT_PULLUP);        // A1

  pinMode(A2,   INPUT_PULLUP);        // A2

  pinMode(A3,   INPUT_PULLUP);        // A3

  pinMode(A4,   INPUT_PULLUP);        // A4

  pinMode(A5,   INPUT_PULLUP);        // A5

#endif

 

#ifdef BASE_NODE_12OUT4IN //----------  Base Node 12 OUT 4 IN ----------

  pinMode(D4,  OUTPUT);             // D4

  pinMode(D5,  OUTPUT);             // D5

  pinMode(D6,  OUTPUT);             // D6

  pinMode(D7,  OUTPUT);             // D7

  pinMode(D8,  OUTPUT);             // D8

  pinMode(D9,  OUTPUT);             // D9

  pinMode(D10, OUTPUT);             // D10

  pinMode(D11, OUTPUT);             // D11

  pinMode(D12, OUTPUT);             // D12

  pinMode(D13, OUTPUT);             // D13

  pinMode(A0,   OUTPUT);             // A0

  pinMode(A1,   OUTPUT);             // A1

 

  pinMode(A2,   INPUT_PULLUP);       // A2

  pinMode(A3,   INPUT_PULLUP);       // A3

  pinMode(A4,   INPUT_PULLUP);       // A4

  pinMode(A5,   INPUT_PULLUP);       // A5

#endif

 

#ifdef BASE_NODE_16IN //----------  Base Node 16 IN ----------

  pinMode(D4,  INPUT_PULLUP);  // D4

  pinMode(D5,  INPUT_PULLUP);  // D5

  pinMode(D6,  INPUT_PULLUP);  // D6

  pinMode(D7,  INPUT_PULLUP);  // D7

  pinMode(D8,  INPUT_PULLUP);  // D8

  pinMode(D9,  INPUT_PULLUP);  // D9

  pinMode(D10, INPUT_PULLUP);  // D10

  pinMode(D11, INPUT_PULLUP);  // D11

 

  pinMode(D12, INPUT_PULLUP);  // D12

  pinMode(D13, INPUT_PULLUP);  // D13

  pinMode(A0,   INPUT_PULLUP);  // A0

  pinMode(A1,   INPUT_PULLUP);  // A1

  pinMode(A2,   INPUT_PULLUP);  // A2

  pinMode(A3,   INPUT_PULLUP);  // A3

  pinMode(A4,   INPUT_PULLUP);  // A4

  pinMode(A5,   INPUT_PULLUP);  // A5

#endif

 

#ifdef BASE_NODE_16OUT //----------  Base Node 16 OUT ----------

  pinMode(D4,  OUTPUT);  // D4

  pinMode(D5,  OUTPUT);  // D5

  pinMode(D6,  OUTPUT);  // D6

  pinMode(D7,  OUTPUT);  // D7

  pinMode(D8,  OUTPUT);  // D8

  pinMode(D9,  OUTPUT);  // D9

  pinMode(D10, OUTPUT);  // D10

  pinMode(D11, OUTPUT);  // D11

 

  pinMode(D12, OUTPUT);  // D12

  pinMode(D13, OUTPUT);  // D13

  pinMode(A0,   OUTPUT);  // A0

  pinMode(A1,   OUTPUT);  // A1

  pinMode(A2,   OUTPUT);  // A2

  pinMode(A3,   OUTPUT);  // A3

  pinMode(A4,   OUTPUT);  // A4

  pinMode(A5,   OUTPUT);  // A5

#endif

 

#ifdef BASE_NODE_RSMC    //----------  Base Node with remote stall motor ----------

  pinMode(D4,  OUTPUT);       // D4  SG1A-R

  pinMode(D5,  OUTPUT);       // D5  SG1A-Y

  pinMode(D6,  OUTPUT);       // D6  SG1A-G

  pinMode(D7,  OUTPUT);       // D7  SG1B-R

  pinMode(D8,  OUTPUT);       // D8  SG1B-Y

  pinMode(D9,  OUTPUT);       // D9  SG2-R

  pinMode(D10, OUTPUT);       // D10 SG2-Y

  pinMode(D11, OUTPUT);       // D11 SG2-G

  pinMode(D12, OUTPUT);       // D12 SG3-R

  pinMode(D13, OUTPUT);       // D13 SG3-Y

  pinMode(A0,   OUTPUT);       // A0  SW1  Turnout throw input to RSMC

 

  pinMode(A1,  INPUT_PULLUP);  // A1  AUX1

  pinMode(A2,  INPUT_PULLUP);  // A2  AUX2

  pinMode(A3,  INPUT_PULLUP);  // A3  TC1  Track Circuit 1

  pinMode(A4,  INPUT_PULLUP);  // A4  TC2  Track Circuit 2

  pinMode(A5,  INPUT_PULLUP);  // A5  OS1  OS Circuit 1

#endif

 

#ifdef BASE_NODE_RSMC_LOCK    //----------  Base Node with remote stall motor ----------

  pinMode(D4,  OUTPUT);       // D4  SG1A-R

  pinMode(D5,  OUTPUT);       // D5  SG1A-Y

  pinMode(D6,  OUTPUT);       // D6  SG1A-G

  pinMode(D7,  OUTPUT);       // D7  SG1B-R

  pinMode(D8,  OUTPUT);       // D8  SG1B-Y

  pinMode(D9,  OUTPUT);       // D9  SG2-R

  pinMode(D10, OUTPUT);       // D10 SG2-Y

  pinMode(D11, OUTPUT);       // D11 SG2-G

  pinMode(D12, OUTPUT);       // D12 SG3-R

  pinMode(D13, OUTPUT);       // D13 SG3-Y

  pinMode(A0,   OUTPUT);       // A0  SW Turnout throw input to RSMC

  pinMode(A1,   OUTPUT);       // A1  SW Lock Control to Switch Lock controller

 

  pinMode(A2,  INPUT_PULLUP);  // A2  Fascia switch control

  pinMode(A3,  INPUT_PULLUP);  // A3  TC1  Track Circuit 1

  pinMode(A4,  INPUT_PULLUP);  // A4  TC2  Track Circuit 2

  pinMode(A5,  INPUT_PULLUP);  // A5  OS1  OS Circuit 1

#endif

 

 

 

#ifdef BASE_NODE_16IN40OUT //----------  Base Node 16 IN  40 out----------

  pinMode(D4,  INPUT_PULLUP);  // D4

  pinMode(D5,  INPUT_PULLUP);  // D5

  pinMode(D6,  INPUT_PULLUP);  // D6

  pinMode(D7,  INPUT_PULLUP);  // D7

  pinMode(D8,  INPUT_PULLUP);  // D8

  pinMode(D9,  INPUT_PULLUP);  // D9

  pinMode(D10, INPUT_PULLUP);  // D10

  pinMode(D11, INPUT_PULLUP);  // D11

 

  pinMode(D12, INPUT_PULLUP);  // D12

  pinMode(D13, INPUT_PULLUP);  // D13

  pinMode(A0,   INPUT_PULLUP);  // A0

  pinMode(A1,   INPUT_PULLUP);  // A1

  pinMode(A2,   INPUT_PULLUP);  // A2

  pinMode(A3,   INPUT_PULLUP);  // A3

  pinMode(A4,   INPUT_PULLUP);  // A4

  pinMode(A5,   INPUT_PULLUP);  // A5

 

  pinMode(A8,    OUTPUT);  // A8

  pinMode(A9,    OUTPUT);  // A9

  pinMode(A10,   OUTPUT);  // A10

  pinMode(A11,   OUTPUT);  // A11

  pinMode(A12,   OUTPUT);  // A12

  pinMode(A13,   OUTPUT);  // A13

  pinMode(A14,   OUTPUT);  // A14

  pinMode(A15,   OUTPUT);  // A15

 

 

  pinMode(D22,  OUTPUT);       // D22

  pinMode(D23,  OUTPUT);       // D23

  pinMode(D24,  OUTPUT);       // D24

  pinMode(D25,  OUTPUT);       // D25

  pinMode(D26,  OUTPUT);       // D26

  pinMode(D27,  OUTPUT);       // D27

  pinMode(D28,  OUTPUT);       // D28

  pinMode(D29,  OUTPUT);       // D29

 

  pinMode(D30,  OUTPUT);       // D30

  pinMode(D31,  OUTPUT);       // D31

  pinMode(D32,  OUTPUT);       // D32

  pinMode(D33,  OUTPUT);       // D33

  pinMode(D34,  OUTPUT);       // D34

  pinMode(D35,  OUTPUT);       // D35

  pinMode(D36,  OUTPUT);       // D36

  pinMode(D37,  OUTPUT);       // D37

 

  pinMode(D38,  OUTPUT);       // D38

  pinMode(D39,  OUTPUT);       // D39

  pinMode(D40,  OUTPUT);       // D40

  pinMode(D41,  OUTPUT);       // D41

  pinMode(D42,  OUTPUT);       // D42

  pinMode(D43,  OUTPUT);       // D43

  pinMode(D44,  OUTPUT);       // D44

  pinMode(D45,  OUTPUT);       // D45

 

  pinMode(D46,  OUTPUT);       // D46

  pinMode(D47,  OUTPUT);       // D47

  pinMode(D48,  OUTPUT);       // D48

  pinMode(D49,  OUTPUT);       // D49

  pinMode(D50,  OUTPUT);       // D50

  pinMode(D51,  OUTPUT);       // D51

  pinMode(D52,  OUTPUT);       // D52

  pinMode(D53,  OUTPUT);       // D53

 

#endif

 

  // Open the monitor port if in debug mode

  //---------------------------------------

  if ((debugMode) || (debugMode1))

  {

    Serial.begin(MON_SPEED);

    while (!Serial);

  }

  else

    Serial.end();

  delay(100);

 

  // Open the CMRInet port

  //----------------------

  Serial1.begin(CMRINET_SPEED);

  while (!Serial1);

  delay(100);

 

  // Call the default initialization routine

  //----------------------------------------

  SetUp_Node();

 

  // Set up I/O Expander (IOX) access if present

  //--------------------------------------------

#ifdef USE_IOX

  Wire.begin();

  Configure_IOX_Ports();

  Setup_IOX_Inputs();

  Setup_IOX_Outputs();

#endif

 

 

 

  // Set the node address

  //---------------------

  SetNodeAddr( nodeID );

  if ((debugMode) || (debugMode1))

  {

    Serial.print("\nCMRI Node "); Serial.print(UA - UA_Offset);

    Serial.print("  Memory Available ");  Serial.println(freeRam());

  }

 

  //---------------------------------------------------

  //  Set all outputs off if Common Anode LEDs are used

  //---------------------------------------------------

  if (Init_CA_Ports_OFF)

  {

    for (i = 0; i < maxOB; i++)

      CMRInet_Buf[i] = 0xFF;

    Unpack_Node_Outputs();

 

#ifdef USE_IOX

    for (i = nOB; i < numIOX_OUT + nOB; i++)

      CMRInet_Buf[i] = 0xFF;

    Unpack_IOX_Outputs();

#endif

  }

 

  // Let the node settle down

  //-------------------------

  delay(100);

 

}  // SETUP

 

 

/**************************************************

*******          Main Loop Code          **********

***************************************************/

void loop()

{

 

  currentMS = millis();

 

  //----------------------------------------------

  //  Check for any messages from the host

  //----------------------------------------------

  switch ( CMRI_Read() )

  {

 

    // No data recveived, ignore

    //--------------------------

    case respNone:

      break;

 

    // Flush input buffer to ETX for various reasons

    //----------------------------------------------

    case respErr:

    //        if (debugMode) Serial.println("Error - Flush to ETX");

    case respIgnore:

      Flush_CMRInet_To_ETX();

      break;

 

    // "I" Initialize Message  HOST -> NODE, set configuration parameters

    //-------------------------------------------------------------------

    case respInit:

      Initialize_cpNode();

      break;

 

    // "P" Poll Message     HOST -> NODE request for input data

    // "R" Receive Message  NODE -> HOST send input port data to host

    //---------------------------------------------------------------

    case respPoll:

      CMRI_Poll_Resp();

      break;

 

    // "T" Transmit (Write) Message  HOST -> NODE, set output bits

    //------------------------------------------------------------

    case respTransmit:

      Unpack_Node_Outputs();

#ifdef USE_IOX

      Unpack_IOX_Outputs();

#endif

      break;

 

    default:

      Flush_CMRInet_To_ETX();

      break;

  }

 

  // Read the input bits and latch for poll response

  //-----------------------------------------------

  Pack_Node_Inputs();   // Pack the onboard inputs

#ifdef USE_IOX

  Pack_IOX_Inputs();    // Pack any inputs off the I2C bus

#endif

 

 

#ifdef DCC_Accessory_Decoder

 

 

  // You MUST call the NmraDcc.process() method frequently from the Arduino loop() function for correct library operation

  Dcc.process();

}

extern void notifyDccAccTurnoutOutput( uint16_t Addr, uint8_t Direction, uint8_t OutputPower ) {

  if ( Addr >= Accessory_Address && Addr < Accessory_Address + DCC_Turnout_Mega_Max) //Controls This_Decoder_Address+16

    digitalWrite( DCC_Turnout_Pins [Addr - Accessory_Address], Direction );

 

 

  for (int i = 0; i <= Addr - Accessory_Address_I2C ; i++) {

    Addr_DCC = DCC_Turnout_Pins_I2C[i] ;

  }

    int k = 0;

    int i = 0;

        Addr_DCC = Addr_DCC - 1;

    for (int j = 0; j <= (max_IOX * 8); j += 8) {

        if (Addr_DCC >= k && Addr_DCC <= k  + 8){

        bitWrite(DCC_I2C[i]  , Addr_DCC   - j, Direction);

          Wire.beginTransmission(IOX_Output_Map[i][I2C_ADDR]);      //// Board Address

          Wire.write(GPIO_REG_BASE + IOX_Output_Map[i][I2C_PORT]);  //// Address port

          Wire.write (DCC_I2C[i]) ;

          Wire.endTransmission();

    }

      k = k + 8;

      i ++;

  }

}

#endif

 
